/*
 * linkedList.c
 * Simple implementation of a linked list. Used for time performance comparisson in building lists on GPUs vs building list on CPUs.
 *
 *  Created on: 23 mei 2015
 *      Author: Henk Mulder
 */

#include "linkedList.h"
#define NULL (LinkedListNode *)0

/**
 * Create a new, initially empty, list.
 */
LinkedList *linkedList_create(){
	LinkedList *result = (LinkedList *)malloc(sizeof(LinkedList));
	result-> head = 0;
	return result;
}

/**
 * Add a node to the list.
 */
void linkedList_add(LinkedList *list, LinkedListNode *node) {
	LinkedListNode *last = list->head;
	node->next = NULL;
	if (last == NULL) {
		list->head = node;
	} else {
		while (last->next != NULL) {
			last = last->next;
		}
		last->next = node;
	}
}

/**
 * Add a node to the list for each key.
 */
void linkedList_addKeys(LinkedList *list, int *keys, int len) {
	int i;
	for (i = 0; i < len; i++) {
		LinkedListNode *node = malloc(
				sizeof(LinkedListNode));
		node->key = keys[i];
		linkedList_add(list, node);
	}
}

/**
 * Get the first occurance in the list with the given key.
 */
LinkedListNode* linkedList_get(LinkedList *list, int key) {
	LinkedListNode *node = list->head;
	while (node != 0 && node->key != key) {
		node = node->next;
	}
	return node;
}

/**
 * Delete al nodes, from and including, the given node.
 * Caller is responsible for handling references to the first node in the chain.
 */
void linkedList_destroyNodeChain(LinkedListNode *node){
	if(node->next != NULL){
		linkedList_destroyNodeChain(node->next);
	}
	free(node);
}

/**
 * Remove the list and all its items.
 */
void linkedList_destroyList(LinkedList *list){
	linkedList_destroyNodeChain(list->head);
	free(list);
}

/**
 * Traverse the list, and print the key of each item to stdout.
 */
void linkedList_printList(LinkedList *list) {
	LinkedListNode *node = list->head;
	while (node != NULL) {
		printf("%d, ", node->key);
		node = node->next;
	}
	printf("\n");
}
